@@ -144,7 +144,7 @@ static void invalid_engines(int i915)
 	igt_assert_eq(__gem_context_set_param(i915, &param), -EFAULT);
 
 	handle = gem_create(i915, 4096 * 3);
-	ptr = gem_mmap__gtt(i915, handle, 4096 * 3, PROT_READ);
+	ptr = gem_mmap__device_coherent(i915, handle, 0, 4096 * 3, PROT_READ);
 	gem_close(i915, handle);
 
 	munmap(ptr, 4096);
@@ -242,6 +242,48 @@ static void idempotent(int i915)
 	gem_context_destroy(i915, p.ctx_id);
 }
 
+static uint32_t batch_create(int i915)
+{
+	const uint32_t bbe = MI_BATCH_BUFFER_END;
+	uint32_t handle = gem_create(i915, 4096);
+
+	gem_write(i915, handle, 0, &bbe, sizeof(bbe));
+	return handle;
+}
+
+static void none(int i915)
+{
+	struct i915_context_param_engines engines = {};
+	struct drm_i915_gem_context_param p = {
+		.ctx_id = gem_context_create(i915),
+		.param = I915_CONTEXT_PARAM_ENGINES,
+		.value = to_user_pointer(&engines),
+		.size = sizeof(engines),
+	};
+
+	gem_context_set_param(i915, &p);
+
+	{
+		struct drm_i915_gem_exec_object2 obj = {
+			.handle = batch_create(i915),
+		};
+		struct drm_i915_gem_execbuffer2 execbuf = {
+			.buffers_ptr = to_user_pointer(&obj),
+			.buffer_count = 1,
+			.rsvd1 = p.ctx_id,
+		};
+
+		for (execbuf.flags = 0;
+		     execbuf.flags <= I915_EXEC_RING_MASK;
+		     execbuf.flags++)
+			igt_assert_eq(__gem_execbuf(i915, &execbuf), -EINVAL);
+
+		gem_close(i915, obj.handle);
+	}
+
+	gem_context_destroy(i915, p.ctx_id);
+}
+
 static void execute_one(int i915)
 {
 	I915_DEFINE_CONTEXT_PARAM_ENGINES(engines , I915_EXEC_RING_MASK + 1);
@@ -405,6 +447,14 @@ static void execute_allforone(int i915)
 	gem_context_destroy(i915, param.ctx_id);
 }
 
+static uint32_t read_result(int timeline, uint32_t *map, int idx)
+{
+	sw_sync_timeline_inc(timeline, 1);
+	while (!READ_ONCE(map[idx]))
+		;
+	return map[idx];
+}
+
 static void independent(int i915)
 {
 #define RCS_TIMESTAMP (0x2000 + 0x358)
@@ -417,12 +467,16 @@ static void independent(int i915)
 		.value = to_user_pointer(&engines),
 		.size = sizeof(engines),
 	};
-	struct drm_i915_gem_exec_object2 results = { .handle = gem_create(i915, 4096) };
+	struct drm_i915_gem_exec_object2 results = {
+		.handle = gem_create(i915, 4096),
+		.offset = gem_aperture_size(i915) / 2,
+	};
 	const uint32_t bbe = MI_BATCH_BUFFER_END;
 	int timeline = sw_sync_timeline_create();
 	uint32_t last, *map;
 
 	igt_require(gen >= 6); /* No per-engine TIMESTAMP on older gen */
+	igt_require(gem_scheduler_enabled(i915));
 
 	{
 		struct drm_i915_gem_execbuffer2 execbuf = {
@@ -438,6 +492,11 @@ static void independent(int i915)
 	memset(&engines, 0, sizeof(engines)); /* All rcs0 */
 	gem_context_set_param(i915, &param);
 
+	gem_set_caching(i915, results.handle, I915_CACHING_CACHED);
+	map = gem_mmap__cpu(i915, results.handle, 0, 4096, PROT_READ);
+	gem_set_domain(i915, results.handle,
+		       I915_GEM_DOMAIN_CPU, I915_GEM_DOMAIN_CPU);
+
 	for (int i = 0; i < I915_EXEC_RING_MASK + 1; i++) {
 		struct drm_i915_gem_exec_object2 obj[2] = {
 			results, /* write hazard lies! */
@@ -472,24 +531,92 @@ static void independent(int i915)
 		gem_close(i915, obj[1].handle);
 		close(execbuf.rsvd2);
 	}
-	close(timeline);
-	gem_sync(i915, results.handle);
 
-	map = gem_mmap__cpu(i915, results.handle, 0, 4096, PROT_READ);
-	gem_set_domain(i915, results.handle, I915_GEM_DOMAIN_CPU, 0);
-	gem_close(i915, results.handle);
-
-	last = map[0];
+	last = read_result(timeline, map, 0);
 	for (int i = 1; i < I915_EXEC_RING_MASK + 1; i++) {
-		igt_assert_f((map[i] - last) > 0,
-			     "Engine instance [%d] executed too late\n", i);
-		last = map[i];
+		uint32_t t = read_result(timeline, map, i);
+		igt_assert_f(t - last > 0,
+			     "Engine instance [%d] executed too late, previous timestamp %08x, now %08x\n",
+			     i, last, t);
+		last = t;
 	}
 	munmap(map, 4096);
 
+	close(timeline);
+	gem_sync(i915, results.handle);
+	gem_close(i915, results.handle);
+
 	gem_context_destroy(i915, param.ctx_id);
 }
 
+static void libapi(int i915)
+{
+	I915_DEFINE_CONTEXT_PARAM_ENGINES(engines, 64) = {};
+	struct drm_i915_gem_context_param p = {
+		.ctx_id = gem_context_create(i915),
+		.param = I915_CONTEXT_PARAM_ENGINES,
+		.value = to_user_pointer(&engines),
+	};
+	const struct intel_execution_engine2 *e;
+	unsigned int count, idx;
+
+	p.size = sizeof(struct i915_context_param_engines);
+	gem_context_set_param(i915, &p);
+
+	/* An empty context should be a short loop */
+	count = 0;
+	for_each_context_engine(i915, p.ctx_id, e)
+		count++;
+	igt_assert_eq(count, 0);
+
+	p.size += sizeof(struct i915_engine_class_instance);
+	engine_class(&engines, 0) = -1;
+	engine_instance(&engines, 0) = -1;
+	gem_context_set_param(i915, &p);
+
+	/* We report all engines from the context, even if invalid/unusable */
+	count = 0;
+	for_each_context_engine(i915, p.ctx_id, e) {
+		igt_assert_eq(e->class, engine_class(&engines, 0));
+		igt_assert_eq(e->instance, engine_instance(&engines, 0));
+		count++;
+	}
+	igt_assert_eq(count, 1);
+
+	/* Check that every known engine can be found from the context map */
+	idx = 0;
+	p.size = sizeof(struct i915_context_param_engines);
+	p.size += sizeof(struct i915_engine_class_instance);
+	for (engine_class(&engines, idx) = 0;
+	     engine_class(&engines, idx) < 16;
+	     engine_class(&engines, idx)++) {
+		for (engine_instance(&engines, idx) = 0;
+		     engine_instance(&engines, idx) < 16;
+		     engine_instance(&engines, idx)++) {
+			if (__gem_context_set_param(i915, &p))
+				break;
+
+			count = 0;
+			for_each_context_engine(i915, p.ctx_id, e) {
+				igt_assert_eq(e->class,
+					      engine_class(&engines, count));
+				igt_assert_eq(e->instance,
+					      engine_instance(&engines, count));
+				count++;
+			}
+			igt_assert_eq(count, idx + 1);
+
+			engines.engines[(idx + 1) % 64] = engines.engines[idx];
+			idx = (idx + 1) % 64;
+
+			p.size = sizeof(struct i915_context_param_engines);
+			p.size += (idx + 1) * sizeof(struct i915_engine_class_instance);
+		}
+	}
+
+	gem_context_destroy(i915, p.ctx_id);
+}
+
 igt_main
 {
 	int i915 = -1;
@@ -500,6 +627,8 @@ igt_main
 
 		gem_require_contexts(i915);
 		igt_require(has_context_engines(i915));
+
+		igt_fork_hang_detector(i915);
 	}
 
 	igt_subtest("invalid-engines")
@@ -508,6 +637,9 @@ igt_main
 	igt_subtest("idempotent")
 		idempotent(i915);
 
+	igt_subtest("none")
+		none(i915);
+
 	igt_subtest("execute-one")
 		execute_one(i915);
 
@@ -519,4 +651,10 @@ igt_main
 
 	igt_subtest("independent")
 		independent(i915);
+
+	igt_subtest("libapi")
+		libapi(i915);
+
+	igt_fixture
+		igt_stop_hang_detector();
 }
