@@ -110,6 +110,9 @@ static void from_mmap(int fd, uint64_t size, enum mode mode)
 	uint64_t max, i;
 	int retry = 2;
 
+	if ((mode & ~RO) == GTT)
+		gem_require_mappable_ggtt(fd);
+
 	/* Worst case is that the kernel has to copy the entire incoming
 	 * reloc[], so double the memory requirements.
 	 */
@@ -258,9 +261,9 @@ static void active(int fd, unsigned engine)
 
 	nengine = 0;
 	if (engine == ALL_ENGINES) {
-		for_each_physical_engine(fd, engine) {
-			if (gem_can_store_dword(fd, engine))
-				engines[nengine++] = engine;
+		for_each_physical_engine(e, fd) {
+			if (gem_can_store_dword(fd, eb_ring(e)))
+				engines[nengine++] = eb_ring(e);
 		}
 	} else {
 		igt_require(gem_has_ring(fd, engine));
@@ -324,6 +327,51 @@ static void active(int fd, unsigned engine)
 	gem_close(fd, obj[0].handle);
 }
 
+static unsigned int offset_in_page(void *addr)
+{
+	return (uintptr_t)addr & 4095;
+}
+
+static void active_spin(int fd, unsigned engine)
+{
+	const uint32_t bbe = MI_BATCH_BUFFER_END;
+	struct drm_i915_gem_relocation_entry reloc;
+	struct drm_i915_gem_exec_object2 obj[2];
+	struct drm_i915_gem_execbuffer2 execbuf;
+	igt_spin_t *spin;
+
+	spin = igt_spin_new(fd,
+			    .engine = engine,
+			    .flags = IGT_SPIN_NO_PREEMPTION);
+
+	memset(obj, 0, sizeof(obj));
+	obj[0] = spin->obj[IGT_SPIN_BATCH];
+	obj[0].relocs_ptr = to_user_pointer(&reloc);
+	obj[0].relocation_count = 1;
+	obj[1].handle = gem_create(fd, 4096);
+	gem_write(fd, obj[1].handle, 0, &bbe, sizeof(bbe));
+
+	memset(&reloc, 0, sizeof(reloc));
+	reloc.presumed_offset = -1;
+	reloc.offset = offset_in_page(spin->condition);
+	reloc.target_handle = obj[0].handle;
+
+	memset(&execbuf, 0, sizeof(execbuf));
+	execbuf.buffers_ptr = to_user_pointer(obj);
+	execbuf.buffer_count = 2;
+	execbuf.flags = engine;
+
+	gem_execbuf(fd, &execbuf);
+	gem_close(fd, obj[1].handle);
+	igt_assert_eq(*spin->condition, spin->cmd_precondition);
+
+	igt_spin_end(spin);
+	gem_sync(fd, spin->handle);
+
+	igt_assert_eq(*spin->condition, obj[0].offset);
+	igt_spin_free(fd, spin);
+}
+
 static bool has_64b_reloc(int fd)
 {
 	return intel_gen(intel_get_drm_devid(fd)) >= 8;
@@ -331,7 +379,8 @@ static bool has_64b_reloc(int fd)
 
 #define NORELOC 1
 #define ACTIVE 2
-#define HANG 4
+#define INTERRUPTIBLE 4
+#define HANG 8
 static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 {
 #define OBJSZ 8192
@@ -342,6 +391,9 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 	const uint32_t bbe = MI_BATCH_BUFFER_END;
 	unsigned int reloc_offset;
 
+	if ((before | after) & I915_GEM_DOMAIN_GTT)
+		gem_require_mappable_ggtt(fd);
+
 	memset(&obj, 0, sizeof(obj));
 	obj.handle = gem_create(fd, OBJSZ);
 	obj.relocs_ptr = to_user_pointer(&reloc);
@@ -684,6 +736,7 @@ igt_main
 		{ "", 0 , true},
 		{ "-noreloc", NORELOC, true },
 		{ "-active", ACTIVE, true },
+		{ "-interruptible", ACTIVE | INTERRUPTIBLE },
 		{ "-hang", ACTIVE | HANG },
 		{ },
 	}, *f;
@@ -711,14 +764,17 @@ igt_main
 					      f->name) {
 					if ((m->before | m->after) & I915_GEM_DOMAIN_WC)
 						igt_require(gem_mmap__has_wc(fd));
-					basic_reloc(fd, m->before, m->after, f->flags);
+					igt_while_interruptible(f->flags & INTERRUPTIBLE)
+						basic_reloc(fd, m->before, m->after, f->flags);
 				}
 			}
 
 			if (!(f->flags & NORELOC)) {
 				igt_subtest_f("%srange%s",
-					      f->basic ? "basic-" : "", f->name)
-					basic_range(fd, f->flags);
+					      f->basic ? "basic-" : "", f->name) {
+					igt_while_interruptible(f->flags & INTERRUPTIBLE)
+						basic_range(fd, f->flags);
+				}
 			}
 
 			igt_fixture {
@@ -749,12 +805,14 @@ igt_main
 	igt_subtest("gpu")
 		from_gpu(fd);
 
-	igt_subtest("active")
+	igt_subtest("basic-active")
 		active(fd, ALL_ENGINES);
 	for (const struct intel_execution_engine *e = intel_execution_engines;
 	     e->name; e++) {
-		igt_subtest_f("active-%s", e->name)
-			active(fd, e->exec_id | e->flags);
+		igt_subtest_f("basic-active-%s", e->name)
+			active(fd, eb_ring(e));
+		igt_subtest_f("basic-spin-%s", e->name)
+			active_spin(fd, eb_ring(e));
 	}
 	igt_fixture
 		close(fd);
