@@ -50,11 +50,7 @@
 #include "igt_sysfs.h"
 #include "igt_debugfs.h"
 #include "igt_device.h"
-
-#define MSR_PKG_CST_CONFIG_CONTROL	0xE2
-/* HSW/BDW: */
-#define  PKG_CST_LIMIT_MASK		0xF
-#define  PKG_CST_LIMIT_C8		0x6
+#include "igt_edid.h"
 
 #define MSR_PC8_RES	0x630
 #define MSR_PC9_RES	0x631
@@ -122,8 +118,6 @@ struct modeset_params lpsp_mode_params;
 struct modeset_params non_lpsp_mode_params;
 struct modeset_params *default_mode_params;
 
-static int8_t *pm_data = NULL;
-
 static int modprobe(const char *driver)
 {
 	return igt_kmod_load(driver, NULL);
@@ -145,15 +139,7 @@ static bool supports_pc8_plus_residencies(void)
 	if (rc != sizeof(val))
 		return false;
 
-	rc = pread(msr_fd, &val, sizeof(uint64_t), MSR_PKG_CST_CONFIG_CONTROL);
-	if (rc != sizeof(val))
-		return false;
-	if ((val & PKG_CST_LIMIT_MASK) < PKG_CST_LIMIT_C8) {
-		igt_info("PKG C-states limited below PC8 by the BIOS\n");
-		return false;
-	}
-
-	return true;
+	return igt_pm_pc8_plus_residencies_enabled(msr_fd);
 }
 
 static uint64_t get_residency(uint32_t type)
@@ -197,6 +183,14 @@ static enum pc8_status get_pc8_status(void)
 		return PC8_DISABLED;
 }
 
+static bool is_suspended(void)
+{
+	if (has_pc8 && !has_runtime_pm)
+		return get_pc8_status() == PC8_ENABLED;
+	else
+		return igt_get_runtime_pm_status() == IGT_RUNTIME_PM_STATUS_SUSPENDED;
+}
+
 static bool wait_for_pc8_status(enum pc8_status status)
 {
 	return igt_wait(get_pc8_status() == status, 10000, 100);
@@ -655,10 +649,10 @@ static bool i2c_read_edid(const char *connector_name, unsigned char *edid)
 	return rc >= 0;
 }
 
-static void format_hex_string(const unsigned char edid[static EDID_LENGTH],
-			      char buf[static EDID_LENGTH * 5 + 1])
+static void format_hex_string(const unsigned char edid[static EDID_BLOCK_SIZE],
+			      char buf[static EDID_BLOCK_SIZE * 5 + 1])
 {
-	for (int i = 0; i < EDID_LENGTH; ++i)
+	for (int i = 0; i < EDID_BLOCK_SIZE; ++i)
 		sprintf(buf+i*5, "0x%02x ", edid[i]);
 }
 
@@ -670,7 +664,7 @@ static void test_i2c(struct mode_set_data *data)
 
 	for (int i = 0; i < data->res->count_connectors; i++) {
 		unsigned char *drm_edid = data->edids[i] ? data->edids[i]->data : NULL;
-		unsigned char i2c_edid[EDID_LENGTH] = {};
+		unsigned char i2c_edid[EDID_BLOCK_SIZE] = {};
 
 		igt_output_t *output = igt_output_from_connector(&display,
 								 data->connectors[i]);
@@ -694,13 +688,13 @@ static void test_i2c(struct mode_set_data *data)
 			continue;
 
 		if (got_i2c_edid && got_drm_edid)
-			edids_equal = (0 == memcmp(drm_edid, i2c_edid, EDID_LENGTH));
+			edids_equal = (0 == memcmp(drm_edid, i2c_edid, EDID_BLOCK_SIZE));
 		else
 			edids_equal = false;
 
 
 		if (!edids_equal) {
-			char buf[5 * EDID_LENGTH + 1];
+			char buf[5 * EDID_BLOCK_SIZE + 1];
 			igt_critical("Detected EDID mismatch on connector %s\n",
 				     connector_name);
 
@@ -754,21 +748,6 @@ static void setup_pc8(void)
 	has_pc8 = true;
 }
 
-static bool dmc_loaded(void)
-{
-	char buf[15];
-	int len;
-
-	len = igt_sysfs_read(debugfs, "i915_dmc_info", buf, sizeof(buf) - 1);
-	if (len < 0)
-	    return true; /* no CSR support, no DMC requirement */
-
-	buf[len] = '\0';
-
-	igt_info("DMC: %s\n", buf);
-	return strstr(buf, "fw loaded: yes");
-}
-
 static void dump_file(int dir, const char *filename)
 {
 	char *contents;
@@ -795,15 +774,15 @@ static bool setup_environment(void)
 
 	init_mode_set_data(&ms_data);
 
-	pm_data = igt_pm_enable_sata_link_power_management();
+	igt_pm_enable_sata_link_power_management();
 
-	has_runtime_pm = igt_setup_runtime_pm();
+	has_runtime_pm = igt_setup_runtime_pm(drm_fd);
 	setup_pc8();
 
 	igt_info("Runtime PM support: %d\n", has_runtime_pm);
 	igt_info("PC8 residency support: %d\n", has_pc8);
 	igt_require(has_runtime_pm);
-	igt_require(dmc_loaded());
+	igt_require(igt_pm_dmc_loaded(debugfs));
 
 out:
 	disable_all_screens(&ms_data);
@@ -820,8 +799,7 @@ static void teardown_environment(void)
 
 	igt_restore_runtime_pm();
 
-	igt_pm_restore_sata_link_power_management(pm_data);
-	free(pm_data);
+	igt_pm_restore_sata_link_power_management();
 
 	fini_mode_set_data(&ms_data);
 
@@ -953,16 +931,25 @@ static void i2c_subtest(void)
 	enable_one_screen(&ms_data);
 }
 
+struct read_entry_elapsed {
+	uint64_t elapsed;
+	char *path;
+} max_read_entry;
+
 static int read_entry(const char *filepath,
 		      const struct stat *info,
 		      const int typeflag,
 		      struct FTW *pathinfo)
 {
+	struct timespec tv = {};
+	uint64_t elapsed;
 	char buf[4096];
 	int fd;
 	int rc;
 
-	igt_assert_f(wait_for_suspended(), "Before opening: %s (%s)\n",
+	igt_nsec_elapsed(&tv);
+
+	igt_assert_f(is_suspended(), "Before opening: %s (%s)\n",
 		     filepath + pathinfo->base, filepath);
 
 	fd = open(filepath, O_RDONLY | O_NONBLOCK);
@@ -977,16 +964,33 @@ static int read_entry(const char *filepath,
 
 	close(fd);
 
+	igt_drop_caches_set(drm_fd, DROP_IDLE); /* flush pm-idle */
 	igt_assert_f(wait_for_suspended(), "After closing: %s (%s)\n",
 		     filepath + pathinfo->base, filepath);
 
+	elapsed = igt_nsec_elapsed(&tv);
+	if (elapsed > max_read_entry.elapsed) {
+		max_read_entry.elapsed = elapsed;
+		free(max_read_entry.path);
+		max_read_entry.path = strdup(filepath);
+	}
+
 	return 0;
 }
 
 static void walk_fs(char *path)
 {
+	max_read_entry.elapsed = 0;
+
 	disable_all_screens_and_wait(&ms_data);
 	nftw(path, read_entry, 20, FTW_PHYS | FTW_MOUNT);
+
+	if (max_read_entry.path) {
+		igt_info("Slowest file + suspend: %s took %.2fms\n",
+			 max_read_entry.path, max_read_entry.elapsed * 1e-6);
+		free(max_read_entry.path);
+		max_read_entry.path = NULL;
+	}
 }
 
 /* This test will probably pass, with a small chance of hanging the machine in
@@ -1037,7 +1041,7 @@ static void debugfs_forcewake_user_subtest(void)
 	igt_assert(wait_for_suspended());
 }
 
-static void gem_mmap_subtest(bool gtt_mmap)
+static void gem_mmap_args(const struct mmap_offset *t)
 {
 	int i;
 	uint32_t handle;
@@ -1049,14 +1053,9 @@ static void gem_mmap_subtest(bool gtt_mmap)
 
 	handle = gem_create(drm_fd, buf_size);
 
-	if (gtt_mmap) {
-		gem_buf = gem_mmap__gtt(drm_fd, handle, buf_size,
-					PROT_READ | PROT_WRITE);
-	}
-	else {
-		gem_buf = gem_mmap__cpu(drm_fd, handle, 0, buf_size, 0);
-	}
-
+	gem_buf = __gem_mmap_offset(drm_fd, handle, 0, buf_size,
+				    PROT_READ | PROT_WRITE, t->type);
+	igt_require(gem_buf);
 
 	for (i = 0; i < buf_size; i++)
 		gem_buf[i] = i & 0xFF;
@@ -1086,13 +1085,9 @@ static void gem_mmap_subtest(bool gtt_mmap)
 	 * suspended. */
 	disable_all_screens_and_wait(&ms_data);
 
-	if (gtt_mmap) {
-		gem_buf = gem_mmap__gtt(drm_fd, handle, buf_size,
-					PROT_READ | PROT_WRITE);
-	}
-	else {
-		gem_buf = gem_mmap__cpu(drm_fd, handle, 0, buf_size, 0);
-	}
+	gem_buf = __gem_mmap_offset(drm_fd, handle, 0, buf_size,
+				    PROT_READ | PROT_WRITE, t->type);
+	igt_require(gem_buf);
 
 	igt_assert(wait_for_suspended());
 
@@ -1265,6 +1260,7 @@ static void gem_execbuf_subtest(void)
 	uint32_t color;
 
 	igt_require_gem(drm_fd);
+	gem_require_blitter(drm_fd);
 
 	/* Create and set data while the device is active. */
 	enable_one_screen_and_wait(&ms_data);
@@ -1564,7 +1560,7 @@ static void fill_igt_fb(struct igt_fb *fb, uint32_t color)
 	int i;
 	uint32_t *ptr;
 
-	ptr = gem_mmap__gtt(drm_fd, fb->gem_handle, fb->size, PROT_WRITE);
+	ptr = gem_mmap__device_coherent(drm_fd, fb->gem_handle, 0, fb->size, PROT_WRITE);
 	for (i = 0; i < fb->size/sizeof(uint32_t); i++)
 		ptr[i] = color;
 	igt_assert(munmap(ptr, fb->size) == 0);
@@ -1634,7 +1630,7 @@ static void cursor_subtest(bool dpms)
 	 * hopefully it has some fences around it. */
 	rc = drmModeRmFB(drm_fd, cursor_fb3.fb_id);
 	igt_assert_eq(rc, 0);
-	gem_set_tiling(drm_fd, cursor_fb3.gem_handle, false, cursor_fb3.strides[0]);
+	__gem_set_tiling(drm_fd, cursor_fb3.gem_handle, false, cursor_fb3.strides[0]);
 	igt_assert(wait_for_suspended());
 
 	rc = drmModeSetCursor(drm_fd, crtc_id, cursor_fb3.gem_handle,
@@ -2043,18 +2039,23 @@ igt_main_args("", long_options, help_str, opt_handler, NULL)
 		modeset_subtest(SCREEN_TYPE_NON_LPSP, 1, WAIT_STATUS | USE_DPMS);
 
 	/* GEM */
-	igt_subtest("gem-mmap-cpu")
-		gem_mmap_subtest(false);
-	igt_subtest("gem-mmap-gtt")
-		gem_mmap_subtest(true);
+	igt_subtest_with_dynamic("gem-mmap-type") {
+		for_each_mmap_offset_type(drm_fd, t) {
+			igt_dynamic_f("%s", t->name)
+				gem_mmap_args(t);
+		}
+	}
+
 	igt_subtest("gem-pread")
 		gem_pread_subtest();
 	igt_subtest("gem-execbuf")
 		gem_execbuf_subtest();
 	igt_subtest("gem-idle")
 		gem_idle_subtest();
-	igt_subtest("gem-evict-pwrite")
+	igt_subtest("gem-evict-pwrite") {
+		gem_require_mappable_ggtt(drm_fd);
 		gem_evict_pwrite_subtest();
+	}
 
 	/* Planes and cursors */
 	igt_subtest("cursor")
@@ -2087,10 +2088,14 @@ igt_main_args("", long_options, help_str, opt_handler, NULL)
 		dpms_mode_unset_subtest(SCREEN_TYPE_LPSP);
 	igt_subtest("dpms-mode-unset-non-lpsp")
 		dpms_mode_unset_subtest(SCREEN_TYPE_NON_LPSP);
-	igt_subtest("fences")
+	igt_subtest("fences") {
+		gem_require_mappable_ggtt(drm_fd);
 		fences_subtest(false);
-	igt_subtest("fences-dpms")
+	}
+	igt_subtest("fences-dpms") {
+		gem_require_mappable_ggtt(drm_fd);
 		fences_subtest(true);
+	}
 
 	/* Modeset stress */
 	igt_subtest("modeset-lpsp-stress")
@@ -2131,10 +2136,14 @@ igt_main_args("", long_options, help_str, opt_handler, NULL)
 		gem_execbuf_stress_subtest(rounds, WAIT_STATUS | WAIT_EXTRA);
 
 	/* power-wake reference tests */
-	igt_subtest("pm-tiling")
+	igt_subtest("pm-tiling") {
+		gem_require_mappable_ggtt(drm_fd);
 		pm_test_tiling();
-	igt_subtest("pm-caching")
+	}
+	igt_subtest("pm-caching") {
+		gem_require_mappable_ggtt(drm_fd);
 		pm_test_caching();
+	}
 
 	igt_fixture
 		teardown_environment();
@@ -2142,6 +2151,8 @@ igt_main_args("", long_options, help_str, opt_handler, NULL)
 	igt_subtest("module-reload") {
 		igt_debug("Reload w/o display\n");
 		igt_i915_driver_unload();
+
+		igt_kmsg(KMSG_INFO "Reloading i915 w/o display\n");
 		igt_assert_eq(igt_i915_driver_load("disable_display=1 mmio_debug=-1"), 0);
 
 		igt_assert(setup_environment());
@@ -2150,6 +2161,8 @@ igt_main_args("", long_options, help_str, opt_handler, NULL)
 
 		igt_debug("Reload as normal\n");
 		igt_i915_driver_unload();
+
+		igt_kmsg(KMSG_INFO "Reloading i915 as normal\n");
 		igt_assert_eq(igt_i915_driver_load("mmio_debug=-1"), 0);
 
 		igt_assert(setup_environment());
